% Projet de l'UE NSR18
% Realisation par Alexandre Houet et Xavier Lagorce
% M2-SAR - ENS Cachan/UPMC

close all
clear all

% Cette partie permet de configurer l'utilisation d'une webcam.
% Elle est commentee, car nous utiliserons des images issues d'un
% appareil photo dans la suite
%----------------------------------------------------------------
% Configuration du peripherique d'acquisition :
%vidobj = videoinput('linuxvideo', 1, 'YUYV_320x240');
%set(vidobj, 'ReturnedColorSpace', 'grayscale');

% Configuration du declencheur pour ne prendre qu'une image
%triggerconfig(vidobj, 'manual');
%vidobj.FramesPerTrigger = 1;
%----------------------------------------------------------------

% Indique si la calibration doit etre effectuee ou non
calib = 0;

% Calibration de la camera si necessaire
if calib == 1,
    % Pour recuperer une image depuis la webcam
     %start(vidobj);
     %trigger(vidobj);
     %wait(vidobj, 3);
     %image = getdata(vidobj, 1);
    % Pour recuperer une image webcam prealablement sauvegardee
     %load('im_calib.mat');
    % Pour lire une photographie
    im_calib = imread('images/mire.jpg');

    % Points traces sur la mire 3D
    %   plan YZ   plan XY   plan XZ
    X = [0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0   2 2 2 2 4 4 4 4 6 6 6 6 8 8 8 8   2 2 2 2 4 4 4 4 6 6 6 6 8 8 8 8];
    Y = [2 4 6 8 2 4 6 8 2 4 6 8 2 4 6 8   2 4 6 8 2 4 6 8 2 4 6 8 2 4 6 8   0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0];
    Z = [2 2 2 2 4 4 4 4 6 6 6 6 8 8 8 8   0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0   2 4 6 8 2 4 6 8 2 4 6 8 2 4 6 8];
    
    % On demande a l'utilisateur de cliquer sur les points de la mire
    figure(1), imagesc(im_calib), colormap(gray), title('Calibration');
    xp = zeros(size(X));
    yp = zeros(size(Y));
    for i=1:size(X,2)
        title(['Calibration point : X=' int2str(X(i)) ' Y=' int2str(Y(i)) ' Z=' int2str(Z(i))]);
        [xi,yi]=ginput(1);
        xp(i)=xi; yp(i)=yi;
    end
    title('Calibration de la camera');

    % On remplit la matrice representative du systeme a resoudre pour
    % trouver la matrice de projection
    Ap = zeros(3*size(xp,2), 12);
    for i=1:size(xp,2)
        Ap(3*(i-1)+1,:) = [0 0 0 0 -X(i) -Y(i) -Z(i) -1 yp(i)*X(i) yp(i)*Y(i) yp(i)*Z(i) yp(i)];
        Ap(3*(i-1)+2,:) = [X(i) Y(i) Z(i) 1 0 0 0 0 -xp(i)*X(i) -xp(i)*Y(i) -xp(i)*Z(i) -xp(i)];
        Ap(3*(i-1)+3,:) = [-yp(i)*X(i) -yp(i)*Y(i) -yp(i)*Z(i) -yp(i) xp(i)*X(i) xp(i)*Y(i) xp(i)*Z(i) xp(i) 0 0 0 0];
    end

    % On resout le systeme grace a une decomposition en valeurs singulieres
    [U,S,V]=svd(Ap);
    p = V(:,end);
    P = zeros(3,4);
    P(1,:)=p(1:4);P(2,:)=p(5:8);P(3,:)=p(9:12);

    % Puis on determine K a partir de P
    [K Rt] = rq(P);
    % normalisation
    K = 1/K(3,3)*K(1:3,1:3);
    
    % On calcule alors l'erreur de reprojection sur P pour verifier
    % que notre matrice est correcte
    xip=xp; yip=yp;
    for i=1:size(X,2)
        pim = P*[X(i) Y(i) Z(i) 1].';
        pim = pim/pim(3);
        xip(i) = pim(1);
        yip(i) = pim(2);
    end
    err2m_p = sqrt(sum((xip-xp).*(xip-xp) + (yip-yp).*(yip-yp)))/length(xp)
    % On dessine la mire
    hold on;
    plot(xip,yip,'b+');
    % On dessine le cube de la mire
    X = [0 8 8 0 0 0 0 0 0 8 8 0   8 8 0  8 8];
    Y = [0 0 8 8 0 0 8 8 0 0 0 0   0 8 8  8 8];
    Z = [0 0 0 0 0 8 8 0 0 0 8 8   8 8 8  8 0];
    xi=X; yi=Y;
    for i=1:size(X,2)
        pim = P*[X(i) Y(i) Z(i) 1].';
        pim = pim/pim(3);
        xi(i) = pim(1);
        yi(i) = pim(2);
    end
    plot(xi(1:12),yi(1:12),'b-');
    plot(xi(13:15),yi(13:15),'b--');
    plot(xi(16:17),yi(16:17),'b--');
else
    % Si on ne refait pas la calibration, il faut charger la matrice K
    % depuis un fichier
    load('camera.mat');
end

% -------------------------------------
% Realite augmentee
% -------------------------------------

% Prise d'une image depuis la webcam
 %start(vidobj);
 %trigger(vidobj);
 %wait(vidobj, 3);
 %image = getdata(vidobj, 1);
% Recuperation d'une image prealablement sauvegardee
 %load('image.mat');
% Recuperation d'une photographie
image = imread('images/image1.jpg');

% Points d'interet du plan sur lequel ajouter notre objet
X = 2*[0 1 2 3 4  0 1 2 3 4  0 1 2 3 4  0 1 2 3 4  0 1 2 3 4  0 1 2 3 4  0 1 2 3 4  0 1 2 3 4  0 1 2 3 4];
Y = 2*[0 0 0 0 0  1 1 1 1 1  2 2 2 2 2  3 3 3 3 3  4 4 4 4 4  5 5 5 5 5  6 6 6 6 6  7 7 7 7 7  8 8 8 8 8];
Z = zeros(1,9*5);

% On demande a l'utilisateur de cliquer sur les points du plan
figure(2), imagesc(image), colormap(gray), title('Reperage du support');
xh = zeros(size(X));
yh = zeros(size(Y));
for i=1:size(X,2)
    title(['Recuperation point : X=' int2str(X(i)) ' Y=' int2str(Y(i)) ' Z=' int2str(Z(i))]);
    [xi,yi]=ginput(1);
    xh(i)=xi; yh(i)=yi;
end
title('Ajout d''un objet dans la scene');

% On remplit la matrice representative du systeme a resoudre pour
% trouver la matrice d'homographie
Ah = zeros(3*size(xh,2), 9);
for i=1:size(xh,2)
    Ah(3*(i-1)+1,:) = [0 0 0 -X(i) -Y(i) -1 yh(i)*X(i) yh(i)*Y(i) yh(i)];
    Ah(3*(i-1)+2,:) = [X(i) Y(i) 1 0 0 0 -xh(i)*X(i) -xh(i)*Y(i) -xh(i)];
    Ah(3*(i-1)+3,:) = [-yh(i)*X(i) -yh(i)*Y(i) -yh(i) xh(i)*X(i) xh(i)*Y(i) xh(i) 0 0 0];
end
% On resout le systeme grace a une decomposition en valeurs singulieres
[U,S,V]=svd(Ah);
h = V(:,end);
H = zeros(3,3);
H(1,:)=h(1:3);H(2,:)=h(4:6);H(3,:)=h(7:9);
H = H/H(end,end);

% Calcul de l'erreur de reprojection sur H
xih=xh; yih=yh;
for i=1:size(X,2)
    pim = H*[X(i) Y(i) 1].';
    pim = pim/pim(3);
    xih(i) = pim(1);
    yih(i) = pim(2);
end
err2m_h = sqrt(sum((xih-xh).*(xih-xh) + (yih-yh).*(yih-yh)))/length(xh)

% On determine [R|t] a partir de H
Pt = inv(K)*H;
Pk = zeros(3,4);
Pk(:,1) = Pt(:,1);
Pk(:,2) = Pt(:,2);
Pk(:,4) = Pt(:,3);
Pk(:,3) = cross(Pk(:,1),Pk(:,2));

% Normalisation de inv(K)*P
% (H connue a un facteur multiplicatif lambda pres)
lambda = nthroot(det(Pk(:,1:3)),4);
Pk(:,1:2)=Pk(:,1:2)./lambda;
Pk(:,3)=Pk(:,3)./(lambda*lambda);
Pk(:,4)=Pk(:,4)./lambda;

% Retour a P
P = K*Pk;

% Trace des valeurs pointees
Xi=X; Yi=Y; Z = 0*ones(size(X));
for i=1:size(X,2)
    pim = P*[X(i) Y(i) Z(i) 1].';
    pim = pim/pim(3);
    Xi(i) = pim(1);
    Yi(i) = pim(2);
end
hold on;
plot(Xi, Yi, 'k+');

% Trace du plan superieur
Z = 0.7*ones(size(X));
for i=1:size(X,2)
    pim = P*[X(i) Y(i) Z(i) 1].';
    pim = pim/pim(3);
    Xi(i) = pim(1);
    Yi(i) = pim(2);
end
hold on;
plot(Xi, Yi, 'r+');

% Trace du plan inferieur
Z = -0.7*ones(size(X));
for i=1:size(X,2)
    pim = P*[X(i) Y(i) Z(i) 1].';
    pim = pim/pim(3);
    Xi(i) = pim(1);
    Yi(i) = pim(2);
end
hold on;
plot(Xi, Yi, 'g+');


% Trace d'un objet 3D :
X = [0 8 8 0 0  0  8  8  0  0  0   4  8 8   4 0  0   4   4 8 8  8];
Y = [0 0 0 0 0 16 16 16 16 16 16  16 16 0   0 0 16  16   0 0 0 16];
Z = 0.7*[0 0 1 1 0  0  0  1  1  0  1 1.5  1 1 1.5 1  1 1.5 1.5 1 0  0];
Xi = zeros(size(X));
Yi = zeros(size(Y));
for i=1:size(X,2)
    pim = P*[X(i) Y(i) Z(i) 1].';
    pim = pim/pim(3);
    Xi(i) = pim(1);
    Yi(i) = pim(2);
end
hold on;
plot(Xi, Yi, 'c');
